import threading
from pycromanager.acquisition.acq_eng_py.main.acq_notification import AcqNotification
from types import GeneratorType

def _axes_to_key(axes):
    """ Turn axes into a hashable key """
    return None if axes is None else frozenset(axes.items())

class AcquisitionFuture:

    def __init__(self, acq, axes_or_axes_list=None):
        """
        :param axes_or_axes_list: a single axes (dictionary) or a list of axes
        """
        self._acq = acq
        self._condition = threading.Condition()
        self._notification_recieved = {}
        self._generator_events = axes_or_axes_list is None
        if not self._generator_events:
            self._add_notifications(axes_or_axes_list)
        self._last_notification = None

    def _add_notifications(self, axes_or_axes_list):
        if isinstance(axes_or_axes_list, dict):
            axes_or_axes_list = [axes_or_axes_list]
        for axes in axes_or_axes_list:
            # TODO maybe unify snap and sequence cause this is confusing
            self._notification_recieved[_axes_to_key(axes)] = {
                AcqNotification.Hardware.PRE_HARDWARE: False,
                AcqNotification.Hardware.PRE_Z_DRIVE: False,
                AcqNotification.Hardware.POST_HARDWARE: False,

                AcqNotification.Camera.PRE_SNAP: False,
                AcqNotification.Camera.PRE_SEQUENCE_STARTED: False,
                AcqNotification.Camera.POST_SNAP: False,
                AcqNotification.Camera.POST_SEQUENCE_STOPPED: False,

                AcqNotification.Image.IMAGE_SAVED: False,
            }


    def _notify(self, notification):
        """
        Called by the internal notification dispatcher in order so that it can check off that the notification was
        received. Want to store this, rather than just waiting around for it, in case the await methods are called
        after the notification has already been sent.
        """
        if notification.milestone == AcqNotification.Acquisition.ACQ_EVENTS_FINISHED or \
            notification.milestone == AcqNotification.Image.DATA_SINK_FINISHED:
            with self._condition:
                self._last_notification = notification
                self._condition.notify_all()
            return
        if isinstance(notification.payload, list):
            keys = [_axes_to_key(ax) for ax in notification.payload]
        else:
            keys = [_axes_to_key(notification.payload)]
        # check if any keys are present in the notification_recieved dict
        if not any([key in self._notification_recieved.keys() for key in keys]):
            return # ignore notifications that aren't relevant to this future
        for key in keys:
            self._notification_recieved[key][notification.milestone] = True
        with self._condition:
            self._last_notification = notification
            self._condition.notify_all()

    def _monitor_axes(self, axes_or_axes_list):
        """
        In the case where the acquisition future is constructed for a Generator, the events to be monitored
        are not known until the generator is run. If user code awaits for an event and that event has already
        passed, the future must be able to check if the event has already passed and return immediately.
        So this function is called by the generator as events are created to add them to the list of events to
        keep track of.

        :param axes_or_axes_list: the axes of the event
        """
        if self._generator_events:
            self._add_notifications(axes_or_axes_list)
        else:
            raise ValueError("This future was not constructed with a generator")

    def await_execution(self, milestone, axes=None):
        """
        Block until the given milestone is executed for the given axes
        :param axes: the axes to wait for
        :param milestone: the milestone to wait for
        """
        key = _axes_to_key(axes)
        if not self._generator_events:
            if key not in self._notification_recieved.keys() or milestone not in self._notification_recieved[key].keys():
                notification = AcqNotification(None, axes, milestone)
                raise ValueError("this future is not expecting a notification for: " + str(notification.to_json()))
        with self._condition:
            while not self._notification_recieved[key][milestone]:
                self._condition.wait()

    def await_image_saved(self, axes=None, return_image=False, return_metadata=False):
        """
        Block until the image with the given axes is saved. Return the image and/or metadata if requested.
        :param axes: the axes of the image to wait for. In the case of None, wait for the next image
        :param return_image: if True, return the image
        :param return_metadata: if True, return the metadata
        """

        if axes is None:
            # wait for the next image to be saved
            with self._condition:
                # wait until something happens
                self._condition.wait()
                while self._last_notification is None or \
                        (not self._last_notification.milestone == AcqNotification.Image.IMAGE_SAVED and \
                        not self._last_notification.milestone == AcqNotification.Image.DATA_SINK_FINISHED):
                    self._condition.wait()
                axes = self._last_notification.payload
        else:
            if isinstance(axes, list):
                keys = [_axes_to_key(ax) for ax in axes]
            else:
                keys = [_axes_to_key(axes)]
            if not self._generator_events and axes is not None:
                # make sure this is a valid axes to wait for associated with this Future
                if any([key not in self._notification_recieved.keys() for key in keys]):
                    raise ValueError("This AcquisitionFuture is not expecting a notification for the given axes")
            # wait until all images are saved
            for key in keys:
                with self._condition:
                    if not self._generator_events:
                        while not self._notification_recieved[key][AcqNotification.Image.IMAGE_SAVED]:
                            self._condition.wait()

        if return_image:
            if isinstance(axes, list):
                return [self._acq.get_dataset().read_image(**ax) for ax in axes]
            else:
                return self._acq.get_dataset().read_image(**axes)
        if return_metadata:
            if isinstance(axes, list):
                return [self._acq.get_dataset().read_metadata(**ax) for ax in axes]
            else:
                return self._acq.get_dataset().read_metadata(**axes)
        if return_image and return_metadata:
            if isinstance(axes, list):
                return [(self._acq.get_dataset().read_image(**ax), self._acq.get_dataset().read_metadata(**ax)) for ax in axes]
            else:
                return self._acq.get_dataset().read_image(**axes), self._acq.get_dataset().read_metadata(**axes)





from pyjavaz.wrappers import JavaObject


class JavaRAMDataStorage:
    """
    A python class that wraps a Java-backend RAM data storage.

    This class maintains an index of which images have been saved, but otherwise routes all calls to the Java
    implementation of the RAM data storage.
    """

    def __init__(self, java_RAM_data_storage):
        self._java_RAM_data_storage = java_RAM_data_storage
        self._index_keys = set()

    def add_index_entry(self, data):
        self._index_keys.add(frozenset(data.items()))

    # def get_channel_names(self):
    #     """
    #     :return: list of channel names (strings)
    #     """
    #     return list(self._channels.keys())

    def get_index_keys(self):
        """
        Return a list of every combination of axes that has a image in this dataset
        """
        frozen_set_list = list(self._index_keys)
        # convert to dict
        return [{axis_name: position for axis_name, position in key} for key in frozen_set_list]

    def is_finished(self) -> bool:
        return self._java_RAM_data_storage.is_finished()

    def has_image(self, channel: int or str, z: int, position: int,
                          time: int, row: int, column: int, **kwargs):
        axes = self._consolidate_axes(channel, z, position, time, row, column, **kwargs)
        key = frozenset(axes.items())
        return key in self._index_keys

    def read_image(self, channel=None, z=None, time=None, position=None, row=None, column=None, **kwargs):
        axes = self._consolidate_axes(channel, z, position, time, row, column, **kwargs)
        key = frozenset(axes.items())
        if key not in self._index_keys:
            return None
        java_hashmap = JavaObject('java.util.HashMap')
        for k, v in axes.items():
            java_hashmap.put(k, v)
        tagged_image = self._java_RAM_data_storage.get_image(java_hashmap)
        pixels = tagged_image.pix
        metadata = tagged_image.tags
        return pixels.reshape(metadata['Height'], metadata['Width'])

    def read_metadata(self, channel=None, z=None, time=None, position=None, row=None, column=None, **kwargs):
        axes = self._consolidate_axes(channel, z, position, time, row, column, **kwargs)
        key = frozenset(axes.items())
        if key not in self._index_keys:
            return None
        java_hashmap = JavaObject('java.util.HashMap')
        for k, v in axes.items():
            java_hashmap.put(k, v)
        tagged_image = self._java_RAM_data_storage.get_image(java_hashmap)
        return tagged_image.tags

    def _consolidate_axes(self, channel: int or str, z: int, position: int,
                          time: int, row: int, column: int, **kwargs):
        """
        Pack axes into a convenient format
        """
        axis_positions = {'channel': channel, 'z': z, 'position': position,
                          'time': time, 'row': row, 'column': column, **kwargs}
        # ignore ones that are None
        axis_positions = {n: axis_positions[n] for n in axis_positions.keys() if axis_positions[n] is not None}
        # for axis_name in axis_positions.keys():
        #     # convert any string-valued axes passed as ints into strings
        #     if self.axes_types[axis_name] == str and type(axis_positions[axis_name]) == int:
        #         axis_positions[axis_name] = self._string_axes_values[axis_name][axis_positions[axis_name]]

        return axis_positions

    def has_new_image(self):
        """
        For datasets currently being acquired, check whether a new image has arrived since this function
        was last called, so that a viewer displaying the data can be updated.
        """
        # pass through to full resolution, since only this is monitored in current implementation
        if not hasattr(self, '_new_image_arrived'):
            return False # pre-initilization
        new = self._new_image_arrived
        self._new_image_arrived = False
        return new

    def as_array(self, axes=None, **kwargs):
        """
        Read all data image data as one big numpy array with last two axes as y, x and preceeding axes depending on data.
        If the data doesn't fully fill out the array (e.g. not every z-slice collected at every time point), zeros will
        be added automatically.

        This function is modeled of the same one in the NDTiff library, but it uses numpy arrays instead of dask arrays
        because the data is already in RAM

        Parameters
        ----------
        axes : list
            list of axes names over which to iterate and merge into a stacked array. The order of axes supplied in this
            list will be the order of the axes of the returned dask array. If None, all axes will be used in PTCZYX order.

        **kwargs :
            names and integer positions of axes on which to slice data
        """
        raise NotImplementedError("This function is not yet implemented")
        # TODO
        # if axes is None:
        #     axes = self.axes.keys()
        #
        # empty_image = np.zeros_like(list(self.images.values())[0])
        # indices = [np.array(self.axes[axis_name]) for axis_name in list(axes)]
        # gridded = np.meshgrid(*indices, indexing='ij')
        # result = np.stack(gridded, axis=-1)
        # flattened = result.reshape((-1, result.shape[-1]))
        # images = []
        # for coord in flattened:
        #     images_key = {key: coord[i] for i, key in enumerate(axes)}
        #     key = frozenset(images_key.items())
        #     if key in self.images.keys():
        #         images.append(self.images[key])
        #     else:
        #         images.append(empty_image)
        # # reshape to Num axes + image size dimensions
        # cube = np.array(images).reshape(tuple(len(i) for i in indices) + empty_image.shape)
        # return cube



